import java.util.ArrayList;
import java.util.Collections;

/**
 * Definicion de la clase AyudaCamino.
 * 
 * @author Sebastian Dieguez
 */
public class AyudaCamino {
	/** El conjunto de Nodos que han sido analizados */
	private ArrayList<Nodo> cerrada = new ArrayList<Nodo>();
	/** El conjunto de nodos que aun no analizamos */
	private ListaOrdenada abierta = new ListaOrdenada();
	/** El tamanio maximo de caminata que estamos dispuestos a aceptar sin rendirnos */
	private int distanciaMaxima = 150;
	/** El conjunto completo de nodos en todo el mapa */
	private Nodo[][] nodos;
	/** TRUE si permitimos movimiento diagonal */
	private boolean permitirDiagonal = false;
	/** array con TRUE donde NO se puede pasar */
	private boolean[][] colisiones = null;

	int mapaAncho = 0;
	int mapaAlto = 0;

	/**
	 * Constructor del ayudante
	 */
	public AyudaCamino(int ancho, int alto, boolean[][] colisiones) {
		this.mapaAncho = ancho;
		this.mapaAlto = alto;
		this.colisiones = colisiones;
		this.nodos = new Nodo[ancho][alto];
		for (int x = 0; x < ancho; x++) {
			for (int y = 0; y < alto; y++) {
				this.nodos[x][y] = new Nodo(x, y);
			}
		}
	}

	/**
	 * Obtiene un camino que va desde un punto origen a un punto destino. esquivando colisiones
	 * dadas.
	 * 
	 * @param colisiones
	 * @param desdeX
	 * @param desdeY
	 * @param hastaX
	 * @param hastaY
	 * @return
	 */
	public Camino obtenerCamino(int desdeX, int desdeY, int hastaX, int hastaY) {
		Camino res = null;
		// Chequeo facil y rapido, si el destino esta bloqueado, no llegaremos alli
		if (this.colisiones[hastaX][hastaY]) {
			return null;
		}
		// estado inicial de A*. El grupo cerrado esta vacio. Solo el inicial
		// el mosaico inicial ya esta en la lista abierta y ya estamos alli
		this.nodos[desdeX][desdeY].costo = 0;
		this.nodos[desdeX][desdeY].profundidad = 0;
		this.cerrada.clear();
		this.abierta.clear();
		this.abierta.add(this.nodos[desdeX][desdeY]);

		this.nodos[hastaX][hastaY].padre = null;

		// BUSCAR *************************

		// mientras no nos excedimos del tamanio maximo de caminata
		int profundidad = 0;
		while ((profundidad < this.distanciaMaxima) && (this.abierta.tamanio() != 0)) {
			// sacamos el primer nodo de la lista abierta, este esta determinado a ser el mas
			// probable proximo paso basado en nuestra heuristica
			Nodo actual = (Nodo) this.abierta.primero();
			if (actual == this.nodos[hastaX][hastaY]) {
				break;
			}

			this.abierta.remove(actual);
			this.cerrada.add(actual);

			// buscar en todos sus vecinos evaluandolos como proximos pasos
			for (int x = -1; x < 2; x++) {
				for (int y = -1; y < 2; y++) {
					// no es un vecino, es el nodo actual
					if ((x == 0) && (y == 0)) {
						continue;
					}

					// si no permitimos movimiento diagonal entonces x o y pueden ser cargados
					if (!this.permitirDiagonal) {
						if ((x != 0) && (y != 0)) {
							continue;
						}
					}

					// determinamos la posicion del vecino y lo evaluamos
					int xp = x + actual.x;
					int yp = y + actual.y;

					if (this.esLugarValido(desdeX, desdeY, xp, yp)) {
						// el costo para llegar a este nodo es el costo del nodo actual mas el costo
						// de movimiento para llegar al nodo. Ojo que la heuristica se usa solo en
						// la lista abierta ordenada
						float costoProxPaso = actual.costo + 1;
						Nodo vecino = this.nodos[xp][yp];

						// si el nuevo costo que encontramos para este noso es mas bajo que lo que
						// era anteriormente entonces se determina que puede haver un mejor camino
						// asi que necesita ser re-evaluado
						if (costoProxPaso < vecino.costo) {
							if (this.abierta.contains(vecino)) {
								this.abierta.remove(vecino);
							}
							if (this.cerrada.contains(vecino)) {
								this.cerrada.remove(vecino);
							}
						}

						// si el nodo no fue descartado aun entonces reiniciamos su costo a nuestro
						// costo actual y lo agregamos como un posible proximo paso (lo mandamos a
						// la lista abierta)
						if (!this.abierta.contains(vecino) && !(this.cerrada.contains(vecino))) {
							vecino.costo = costoProxPaso;
							vecino.heuristica = this.getCostoHeuristico(xp, yp, hastaX, hastaY);
							profundidad = Math.max(profundidad, vecino.setPadre(actual));
							this.abierta.add(vecino);
						}
					}
				}
			}
		}

		// BUSCAR *************************

		// como agotamos la busqueda y no hay un camino, retornamos null
		if (this.nodos[hastaX][hastaY].padre == null) {
			return null;
		}
		// En este punto definitivamente encontramos un camino asi que podemos usar las referencias
		// de padre para encontrar el camino desde el objetivo de vuelta hasta la posicion inicial y
		// guardamos los nodos del camino.
		res = new Camino();
		Nodo objetivo = this.nodos[hastaX][hastaY];
		while (objetivo != this.nodos[desdeX][desdeY]) {
			res.anteponerPunto(objetivo.x, objetivo.y);
			objetivo = objetivo.padre;
		}
		res.anteponerPunto(desdeX, desdeY);

		// eso es todo, tenemos nuestro camino
		return res;
	}

	/**
	 * Calcula el costo con una heuristica Ecludiana
	 * 
	 * @param actualX
	 *            Coordenada X del punto actual
	 * @param actualY
	 *            Coordenada Y del punto actual
	 * @param hastaX
	 *            Coordenada X del punto objetivo
	 * @param hastaY
	 *            Coordenada Y del punto objetivo
	 * @return el valor de esta euristica
	 */
	private float getCostoHeuristico(int actualX, int actualY, int hastaX, int hastaY) {
		float dx = hastaX - actualX;
		float dy = hastaY - actualY;

		float resultado = (float) (Math.sqrt((dx * dx) + (dy * dy)));

		return resultado;
	}

	/**
	 * Comprueba si el lugar dado es valido para un movedor indicado
	 * 
	 * @param sx
	 *            La coordenada x inicial
	 * @param sy
	 *            La coordenada y inicial
	 * @param x
	 *            La coordenada x del lugar a comprobar
	 * @param y
	 *            La coordenada y del lugar a comprobar
	 * @return TRUE si el lugar es valido para el movedor dado
	 */
	private boolean esLugarValido(int sx, int sy, int x, int y) {
		boolean invalido = (x < 0) || (y < 0) || (x >= this.mapaAncho) || (y >= this.mapaAlto);

		if ((!invalido) && ((sx != x) || (sy != y))) {
			invalido = this.colisiones[x][y];
		}

		return !invalido;
	}

	/**
	 * Una simple lista ordenada
	 * 
	 * @author Sebastian Dieguez
	 */
	private class ListaOrdenada {
		/** La lista de elementos */
		private ArrayList<Nodo> lista = new ArrayList<Nodo>();

		/**
		 * Obtiene el primer Nodo de la lista
		 * 
		 * @return El primer Nodo de la lista
		 */
		public Nodo primero() {
			return this.lista.get(0);
		}

		/**
		 * Vaciar la lista
		 */
		public void clear() {
			this.lista.clear();
		}

		/**
		 * Agrega un elemento a la lista - provoca el re-ordenamiento
		 * 
		 * @param nodo
		 *            El elemento a agregar
		 */
		public void add(Nodo nodo) {
			this.lista.add(nodo);
			Collections.sort(this.lista);
		}

		/**
		 * Borra un elemento de la lista
		 * 
		 * @param nodo
		 *            El elemento a remover
		 */
		public void remove(Nodo nodo) {
			this.lista.remove(nodo);
		}

		/**
		 * Obtiene el numero de elementos de la lista
		 * 
		 * @return El numero de elementos de la lista
		 */
		public int tamanio() {
			return this.lista.size();
		}

		/**
		 * Comprueba si un elemento esta en la lista
		 * 
		 * @param nodo
		 *            El elemento a buscar
		 * @return TRUE si el elemento esta en la lista
		 */
		public boolean contains(Nodo nodo) {
			return this.lista.contains(nodo);
		}
	}

	/**
	 * Un Nodo simple en el camino de busqueda
	 * 
	 * @author Sebastian Dieguez
	 */
	private class Nodo implements Comparable<Nodo> {
		/** La coordenada x del nodo */
		private int x;
		/** La coordenada y del nodo */
		private int y;
		/** El costo de camino de este nodo */
		private float costo;
		/** El padre de este nodo, gracias al padre llegamos a este nodo */
		private Nodo padre;
		/** El costo heuristico de este nodo */
		private float heuristica;
		/** El indice de profundidad de este nodo, es la posicion en la caminata */
		private int profundidad;

		/**
		 * Crea un nodo
		 * 
		 * @param x
		 *            La coordenada x del nodo
		 * @param y
		 *            La coordenada y del nodo
		 */
		public Nodo(int x, int y) {
			this.x = x;
			this.y = y;
		}

		/**
		 * Carga el Padre de este nodo
		 * 
		 * @param padre
		 *            El padre de este nodo, gracias al padre llegamos a este nodo
		 * @return La profundidad que alcanzamos buscando en esta caminara
		 */
		public int setPadre(Nodo padre) {
			this.profundidad = padre.profundidad + 1;
			this.padre = padre;

			return this.profundidad;
		}

		/**
		 * @see Comparable#compareTo(Nodo)
		 */
		@Override
		public int compareTo(Nodo otro) {

			float costoTotal = this.heuristica + this.costo;
			float otroCostoTotal = otro.heuristica + otro.costo;

			if (costoTotal < otroCostoTotal) {
				return -1;
			} else if (costoTotal > otroCostoTotal) {
				return 1;
			} else {
				return 0;
			}
		}
	}
}
