import greenfoot.Greenfoot;
import greenfoot.World;

import java.util.ArrayList;
import java.util.List;

/**
 * Definicion de la clase GoblinFighterGame.
 * 
 * @author Sebastian Dieguez
 */
public class WarInTheDeepGame extends World {

	public static final int MAPA_ANCHO = 15;
	public static final int MAPA_ALTO = 10;
	public static final int MAPA_SPRITE = 64;

	/**
	 * Configura el mapa de colisiones para considerar las tumbas (pj muertos)
	 */
	private static final boolean PISAR_TUMBAS = true;

	/**
	 * Permite esperar un turno para mostrar el mensaje de finalizacion del juego
	 */
	public boolean CASI_TERMINA = false;

	// Tablero de puntajes
	private int PUNTOS_BUENOS = 0;
	private int PUNTOS_MALOS = 0;

	/**
	 * Cuadro actual del juego
	 */
	private long cuadroJuego = 0;

	/**
	 * Inicializa el mundo
	 */
	public WarInTheDeepGame() {
		super(MAPA_ANCHO, MAPA_ALTO, MAPA_SPRITE, true);
		this.setBackground("fondo.jpg");
		this.setPaintOrder(EtiquetaPluma.class, GoblinTroll.class);
		this.reiniciarTablero(true);
		int mapa = AyudaVistaEntrada.elegirMapa();

		this.cargarMapaEjemplo(mapa);
	}

	/**
	 * Carga el escenario con materiales (0.. 6)
	 */
	public void cargarMapaEjemplo(int numNivel) {

		this.removeObjects(this.getObjects(null));
		this.reiniciarTablero(true);

		// Definimos algunos niveles
		List<String[]> niveles = AyudaNiveles.obtenerNiveles();

		String[] cargaNivel = niveles.get(numNivel);
		boolean cargarPj = true;
		String nombreNivel = "Nivel desconocido";

		int y = 0;
		for (String renglon : cargaNivel) {
			if (y == 0) {
				nombreNivel = renglon;
				cargarPj = AyudaVistaEntrada.decidirCargarPj(nombreNivel);
			} else {
				for (int x = 0; x < renglon.length(); x++) {
					char charAt = renglon.charAt(x);
					// Cargo un objeto pero la coordenada y le resto 1 por el renglon de Nombre del mapa
					this.cargaObjeto(charAt, x, y - 1, cargarPj);
				}
			}
			y++;
		}
	}

	/**
	 * Carga un objeto en el escenario dependiendo de el caracter recibido en el renglon
	 * 
	 * @param c el caracter representando un objeto
	 * @param x la coordenada x del objeto nuevo
	 * @param y la coordenada y del objeto nuevo
	 * @param cargarPj TRUE si se desean cargar personajes
	 */
	private void cargaObjeto(char c, int x, int y, boolean cargarPj) {
		switch (c) {
			case ' ':
				// NADA
				break;
			case '#':
			case '|':
			case '!':
			case '_':
			case 'L':
			case 'W':
				this.addObject(new Muro(c), x, y); // Tipos de Muro
				break;
			case '&':
				if (cargarPj) {
					this.addObject(new EnanoGuerrero(), x, y);
				}
				break;
			case '%':
				if (cargarPj) {
					this.addObject(new EnanoArquero(), x, y);
				}
				break;
			case '$':
				if (cargarPj) {
					this.addObject(new EnanoMonje(), x, y);
				}
				break;
			case 'G':
				if (cargarPj) {
					this.addObject(new GoblinGuerrero(), x, y);
				}
				break;
			case 'A':
				if (cargarPj) {
					this.addObject(new GoblinArquero(), x, y);
				}
				break;
			case 'T':
				if (cargarPj) {
					this.addObject(new GoblinTroll(), x, y);
				}
				break;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see greenfoot.World#act()
	 */
	@Override
	public void act() {
		this.prepararPersonajes();
		this.comprobarCaminos();
		this.ejecutarAcciones();
		this.estadoDelJuego();
		this.cuadroJuego++;
	}

	/**
	 * Recopila los puntajes y decide si hubo un ganador o si el juego termino
	 */
	private void estadoDelJuego() {
		List<Personaje> buenos = this.obtenerBuenosMuchachos(true);
		List<Personaje> malos = this.obtenerChicosMalos(true);

		// Vamos a esperar que todos celebren
		boolean ganaMalo = true;
		if (malos.size() == 0) {
			ganaMalo = false;
		}
		boolean ganaBueno = true;
		if (buenos.size() == 0) {
			ganaBueno = false;
		}
		for (Personaje pj : buenos) {
			if (!pj.GANADOR) {
				ganaBueno = false;
				break;
			}
		}
		for (Personaje pj : malos) {
			if (!pj.GANADOR) {
				ganaMalo = false;
				break;
			}
		}

		// Hacemos esperar un turno para que se muestren las plumas de GANE
		if (!this.CASI_TERMINA) {
			if (ganaBueno || ganaMalo) {
				this.CASI_TERMINA = true;
			}
		} else {
			if (ganaBueno || ganaMalo) {
				StringBuilder sb = new StringBuilder();
				if (ganaMalo) {
					sb.append("GANARON LOS CHICOS MALOS!!!\n\n");
				}
				if (ganaBueno) {
					sb.append("GANARON LOS BUENOS MUCHACHOS!!!\n\n");
				}

				sb.append("LOS BUENOS MUCHACHOS: " + this.PUNTOS_BUENOS + " pts. (" + buenos.size() + " vivos)\n");
				sb.append("LOS CHICOS MALOS: " + this.PUNTOS_MALOS + " pts. (" + malos.size() + " vivos)\n\n");

				sb.append("Deseas borrar los personajes (dejando las paredes)?\n\n");

				boolean decision = AyudaVistaEntrada.decidirBorrarAlFinal(sb.toString());
				// Si eligio borrar los personajes
				if (decision) {
					this.reiniciarTablero(false);
				}
				// Detenemos el juego
				Greenfoot.stop();
			}
		}

	}

	public void reiniciarTablero(boolean paredes) {
		this.PUNTOS_BUENOS = 0;
		this.PUNTOS_MALOS = 0;
		this.cuadroJuego = 0;
		this.CASI_TERMINA = false;

		this.removeObjects(this.obtenerBuenosMuchachos());
		this.removeObjects(this.obtenerChicosMalos());
		this.removeObjects(this.getObjects(EtiquetaPluma.class));
		if (paredes) {
			this.removeObjects(this.getObjects(Muro.class));
		}

	}

	/**
	 * Ejecuta la accion que cada personaje desea realizar
	 */
	private void ejecutarAcciones() {
		List<Personaje> buenos = this.obtenerBuenosMuchachos(true);
		List<Personaje> malos = this.obtenerChicosMalos(true);

		for (Personaje cm : malos) {
			cm.ejecutarAcciones();
		}
		for (Personaje bm : buenos) {
			bm.ejecutarAcciones();
		}
	}

	/**
	 * Decide si el personaje va a golpear, caminar, etc.
	 */
	private void prepararPersonajes() {
		List<Personaje> buenos = this.obtenerBuenosMuchachos(true);
		List<Personaje> malos = this.obtenerChicosMalos(true);

		for (Personaje cm : malos) {
			cm.procesarEstado();
		}
		for (Personaje bm : buenos) {
			bm.procesarEstado();
		}
	}

	/**
	 * Decide para cada personaje, que camino debe tomar
	 */
	private void comprobarCaminos() {
		// Cada 3 turnos
		if ((this.cuadroJuego % AyudaCalculos.tirarDado(5)) == 0) {

			List<Personaje> buenos = this.obtenerBuenosMuchachos(true);
			List<Personaje> malos = this.obtenerChicosMalos(true);
			AyudaCamino caminiante = new AyudaCamino(MAPA_ANCHO, MAPA_ALTO, this.obtenerColisiones());

			// Actualizamos los caminos para los chicos malos
			for (Personaje cm : malos) {
				this.obtenerYCargarCamino(caminiante, cm);
			}

			// Actualizamos los caminos para los buenos muchachos
			for (Personaje bm : buenos) {
				this.obtenerYCargarCamino(caminiante, bm);
			}
		}
	}

	/**
	 * Cargamos en el personaje un nuevo camino
	 * 
	 * @param caminiante
	 * @param pj
	 */
	private void obtenerYCargarCamino(AyudaCamino caminiante, Personaje pj) {
		if (pj.getPjObjetivo() != null) {

			// Buscamos un lugar vacio a su lado
			Punto nuevoPunto = this.obtenerPuntoVacioCerca(pj.getPjObjetivo().getX(), pj.getPjObjetivo().getY());

			Camino nuevoCamino = caminiante.obtenerCamino(pj.getX(), pj.getY(), nuevoPunto.getX(), nuevoPunto.getY());
			pj.setCaminoObjetivo(nuevoCamino);
			if (nuevoCamino == null) {
				pj.estado = 0; // Psivo, no tiene como llegar
				pj.setPjObjetivo(null);
			}
		} else {
			pj.estado = 0; // Psivo, no tiene a donde ir
		}
	}

	/**
	 * Dada unas coordenadas, obtiene un punto vacio vecino revisando los nueve vecinos. Sino devuelve el mismo punto que se entrego.
	 * 
	 * @param x coordenada x del punto
	 * @param y coordenada y del punto
	 * @return un punto cercano o el mismo punto
	 */
	public Punto obtenerPuntoVacioCerca(int x, int y) {
		Punto res = new Punto(x, y);
		if (x > 1) {
			if ((this.getObjectsAt(x - 1, y - 1, null).size() == 0) && (y > 1)) {
				res = new Punto(x - 1, y - 1);
			}
			if (this.getObjectsAt(x - 1, y, null).size() == 0) {
				res = new Punto(x - 1, y);
			}
			if ((this.getObjectsAt(x - 1, y + 1, null).size() == 0) && (y < (MAPA_ALTO - 1))) {
				res = new Punto(x - 1, y + 1);
			}
		}
		if (x < (MAPA_ANCHO - 1)) {
			if ((this.getObjectsAt(x + 1, y - 1, null).size() == 0) && (y > 1)) {
				res = new Punto(x + 1, y - 1);
			}
			if (this.getObjectsAt(x + 1, y, null).size() == 0) {
				res = new Punto(x + 1, y);
			}
			if ((this.getObjectsAt(x + 1, y + 1, null).size() == 0) && (y < (MAPA_ALTO - 1))) {
				res = new Punto(x + 1, y + 1);
			}
		}
		if ((this.getObjectsAt(x, y - 1, null).size() == 0) && (y > 1)) {
			res = new Punto(x, y - 1);
		}
		if ((this.getObjectsAt(x, y + 1, null).size() == 0) && (y < (MAPA_ALTO - 1))) {
			res = new Punto(x, y + 1);
		}

		return res;
	}

	@SuppressWarnings("unchecked")
	private List<ChicosMalos> obtenerChicosMalos() {
		List<ChicosMalos> res = new ArrayList<ChicosMalos>();

		res.addAll(this.getObjects(ChicosMalos.class));

		return res;
	}

	public List<Personaje> obtenerChicosMalos(boolean vivos) {
		List<Personaje> res = new ArrayList<Personaje>();

		List<ChicosMalos> lista = this.obtenerChicosMalos();
		for (ChicosMalos cm : lista) {
			if (cm.SALUD > 0) {
				if (vivos) {
					res.add(cm);
				}
			} else {
				if (!vivos) {
					res.add(cm);
				}
			}
		}
		return res;
	}

	@SuppressWarnings("unchecked")
	private List<BuenosMuchachos> obtenerBuenosMuchachos() {
		List<BuenosMuchachos> res = new ArrayList<BuenosMuchachos>();

		res.addAll(this.getObjects(BuenosMuchachos.class));

		return res;
	}

	public List<Personaje> obtenerBuenosMuchachos(boolean vivos) {
		List<Personaje> res = new ArrayList<Personaje>();
		List<BuenosMuchachos> lista = this.obtenerBuenosMuchachos();
		for (BuenosMuchachos bm : lista) {
			if (bm.SALUD > 0) {
				if (vivos) {
					res.add(bm);
				}
			} else {
				if (!vivos) {
					res.add(bm);
				}
			}
		}
		return res;
	}

	@SuppressWarnings("unchecked")
	public List<Muro> obtenerMuros() {
		List<Muro> res = new ArrayList<Muro>();

		res.addAll(this.getObjects(Muro.class));

		return res;
	}

	/**
	 * Agrega un punto (por danio) al equipo de los buenos
	 */
	public void sumarPuntoBuenos(int danio) {
		this.PUNTOS_BUENOS += danio;
	}

	/**
	 * Agrega un punto (por danio) al equipo de los malos
	 */
	public void sumarPuntoMalos(int danio) {
		this.PUNTOS_MALOS += danio;
	}

	/**
	 * Obtiene un arreglo bidimensional con los elementos bloqueantes (true) o traspasables (false)
	 * 
	 * @return un arreglo con el mapa representando (en TRUE) las colisiones posibles
	 */
	public boolean[][] obtenerColisiones() {
		boolean[][] res = new boolean[MAPA_ANCHO][MAPA_ALTO];

		// Inicializamos el arreglo con FALSE (no colisionar)
		for (int y = 0; y < MAPA_ALTO; y++) {
			for (int x = 0; x < MAPA_ANCHO; x++) {
				res[x][y] = false;
			}
		}

		// Buscamos los objetos (muros y PJ vivos)
		List<Muro> paredes = this.obtenerMuros();
		// Las paredes no se pueden traspasar
		for (Muro pared : paredes) {
			res[pared.getX()][pared.getY()] = true;
		}

		if (!PISAR_TUMBAS) {
			// Las TUMBAS no se pisan
			List<Personaje> listaTumbas = new ArrayList<Personaje>();

			listaTumbas.addAll(this.obtenerBuenosMuchachos(false));
			listaTumbas.addAll(this.obtenerChicosMalos(false));
			for (Personaje pj : listaTumbas) {
				// Si esta muerto
				if (pj.SALUD <= 0) {
					res[pj.getX()][pj.getY()] = true;
				}
			}
		}

		return res;
	}

}
